/*************************************************************************/
/**---------------------------------------------------------------------**/
/**                            AUDIO DRIVER                             **/
/**---------------------------------------------------------------------**/
/*************************************************************************/

/* Sample formats */
#define MODE_16BIT 0
#define MODE_8BIT  1
#define MODE_ADPCM 3

/* Volume 0 is max, higher is lower */
#define VOL_MAX    0x00
#define VOL_MIN    0xff

/* Pan is 5 bit sign-value (it seems) */
#define PAN_LEFT   0x1f
#define PAN_RIGHT  0x0f
#define PAN_CENTER 0x00

/* Addresses */
#define AICA  0xA0700000
#define BUFL  0xA09E0000
#define BUFR  0xA09C0000
#define ABUFL   0x1E0000
#define ABUFR   0x1C0000

static struct audio_param_st {
	int freq;	// Sample rate
	int nch;	// Number of channels
	int format;	// 0: 16 bits, signed, 1: 8 bit, unsigned
	int vol;	// Volume attenuation (log, 8 bits)
	int pos;	// buffer position
	int start;	// delayed start flag
} audio_param;

unsigned int audio_time=0,samples=0;

/*--------------------------------------------------*/

void aica_reset()
{
  int i, j;
  volatile int del;
  volatile unsigned int *hwptr = (unsigned int *)AICA;

  *(volatile unsigned int *)(AICA+0x2800) = 0;

  /* Reset all 64 channels to a silent state */
  for (del=0;del<100;del++);
  for(i=0; i<64; i++) {
    hwptr[0] = 0x8000;
    hwptr[5] = 0x1f;
    hwptr[1] = 0;
    hwptr[2] = 0;
    hwptr[3] = 0;
    hwptr[4] = 0;
    for(j=6; j<32; j++)
      hwptr[j] = 0;
    hwptr += 32;
  }
  *(volatile unsigned int *)(AICA+0x2800) = 15;
  for (del=0;del<100;del++);
}

/*--------------------------------------------------*/

void audio_setup()
{
  volatile unsigned int *hwptrl = (unsigned int *)(AICA+(63<<7));
  volatile unsigned int *hwptrr = (unsigned int *)(AICA+(62<<7));
  unsigned long freq_lo, freq_base = 5644800;
  int freq_hi = 7,freq;

  /* Set sample format and buffer address */
  hwptrl[0] = 0x4200 | (MODE_16BIT<<7) | (ABUFL>>16);
  hwptrl[1] = ABUFL & 0xffff;
  hwptrr[0] = 0x4200 | (MODE_16BIT<<7) | (ABUFR>>16);
  hwptrr[1] = ABUFR & 0xffff;
  /* sample start & end */
  hwptrl[2] =0;
  hwptrr[2] =0;
  hwptrl[3] = 0xffff;
  hwptrr[3] = 0xffff;
  /* Need to convert frequency to floating point format
     (freq_hi is exponent, freq_lo is mantissa).
     Formula is  freq = 44100*2^freq_hi*(1+freq_lo/1024)  */
  freq=audio_param.freq;
  while(freq < freq_base && freq_hi > -8) {
    freq_base >>= 1;
    --freq_hi;
  }
  freq_lo = (freq<<10)/freq_base;
  /* Write resulting values */
  hwptrl[6] = (freq_hi<<11)|(freq_lo&1023);
  hwptrr[6] = (freq_hi<<11)|(freq_lo&1023);
  /* Set volume, pan, and some other stuff */

  if (audio_param.nch == 2) {
  	hwptrl[9] = PAN_LEFT + (0xf<<8);
  	hwptrr[9] = PAN_RIGHT + (0xf<<8);
  } else hwptrl[9] = PAN_CENTER + (0xf<<8);

  hwptrl[10]= ((audio_param.vol & 0xff)<<8) + 0x24;
  hwptrr[10]= ((audio_param.vol & 0xff)<<8) + 0x24;

  hwptrl[4] = 0x1f;
  hwptrr[4] = 0x1f;
  /* Enable playback */
  //hwptr[0] |= 0xc000;
}

/*--------------------------------------------------*/

void audio_start()
{
  volatile unsigned int *hwptrl = (unsigned int *)(AICA+(63<<7));
  volatile unsigned int *hwptrr = (unsigned int *)(AICA+(62<<7));

  /* enable playback */

  if(audio_param.nch == 2) {
  	//cli
  	hwptrl[0] |= 0xc000;
  	hwptrr[0] |= 0xc000;
  	//sti
  } else {
  	hwptrl[0] |= 0xc000;
  }
  *(volatile unsigned int *)(AICA+0x280c)=63<<8;// read channel pointer
  audio_param.pos=*(volatile unsigned int *)(AICA+0x2814);
  audio_time=samples=0;
}

/*--------------------------------------------------*/

void audio_stop(int fast)
{
  volatile unsigned int *hwptrl = (unsigned int *)(AICA+(63<<7));
  volatile unsigned int *hwptrr = (unsigned int *)(AICA+(62<<7));
  volatile int del;

  if (!fast) {
  	if ((hwptrl[0]&0xc000)==0xc000) {
	  	//Wait until the buffer is empty
	  	*(volatile unsigned int *)(AICA+0x280c)=63<<8;
  		for (del=0;del<20;del++);	// This delay is IMPORTANT
	  	while (audio_param.pos != *(volatile unsigned int *)(AICA+0x2814));
	}
  }

  /* Disable playback */
  hwptrl[0] = (hwptrl[0] & ~0x4000) | 0x8000;
  if(audio_param.nch == 2) {
  	hwptrr[0] = (hwptrr[0] & ~0x4000) | 0x8000;
  }
}

/*--------------------------------------------------*/

void audio_put_sample(int datal,int datar)
{
	unsigned int pos;
        static int evenl,evenr;
        volatile int del;

        if (audio_param.format) {
        	datal-=128; datal<<=8;
        	datar-=128; datar<<=8;
        }

	pos=(audio_param.pos+1) & 0xffff;
	
	*(volatile unsigned int *)(AICA+0x280c)=63<<8;
	for (del=0;del<20;del++);	// This delay is IMPORTANT
	while (pos == *(volatile unsigned int *)(AICA+0x2814)); // Schedule here

	if(audio_param.pos&1) {
		((int *)BUFL)[audio_param.pos>>1]=(datal<<16)|(evenl & 0xffff);
	} else {
		evenl=datal;
	}
	if (audio_param.nch == 2) {
		if(audio_param.pos&1) {
			((int *)BUFR)[audio_param.pos>>1]=(datar<<16)|(evenr & 0xffff);
		} else {
			evenr=datar;
		}		
	}
	audio_param.pos=pos;
	if (audio_param.start) {
		audio_param.start=0;
		audio_start();
	}
	samples++;
	if (samples==audio_param.freq) { samples=0; audio_time++; }
}

void silence()
{
	int i;
	for (i=0;i<32768;i++) {
		((int *)BUFL)[i]=0;
		((int *)BUFR)[i]=0;
	}
}

/**********-------------------------------------------***********/

void DAC_init()
{
	/* Pull RESET low on ARM */
	*((volatile unsigned long *)(void *)0xa0702c00) |= 1;
	aica_reset();
	audio_param.freq=44100;
	audio_param.nch=2;
	audio_param.format=0;
	audio_param.vol=6;
	audio_param.start=1;
	silence();
	audio_setup();	
}

void DAC_set(int rate, int nch)
{
	audio_param.freq=rate;
	audio_param.nch=nch;
	audio_setup();
}

void DAC_write(unsigned int *buffer, int nsamples)
{
	for (;nsamples;nsamples--) {
		audio_put_sample(*buffer&0xffff,*buffer>>16);
		buffer++;
	}
}
